@@ -2,7 +2,9 @@ use std::collections::BTreeMap;
 
 use crate::ast;
 use crate::error::Error;
-use crate::instructions::{Instruction, Instructions};
+use crate::instructions::{
+    Instruction, Instructions, LOOP_FLAG_RECURSIVE, LOOP_FLAG_WITH_LOOP_VAR,
+};
 use crate::tokens::Span;
 use crate::utils::matches;
 use crate::value::Value;
@@ -58,8 +60,15 @@ impl<'source> Compiler<'source> {
     }
 
     /// Starts a for loop
-    pub fn start_for_loop(&mut self, with_loop_var: bool) {
-        self.add(Instruction::PushLoop(with_loop_var));
+    pub fn start_for_loop(&mut self, with_loop_var: bool, recursive: bool) {
+        let mut flags = 0;
+        if with_loop_var {
+            flags |= LOOP_FLAG_WITH_LOOP_VAR;
+        }
+        if recursive {
+            flags |= LOOP_FLAG_RECURSIVE;
+        }
+        self.add(Instruction::PushLoop(flags));
         let iter_instr = self.add(Instruction::Iterate(!0));
         self.pending_block.push(PendingBlock::Loop(iter_instr));
     }
@@ -165,6 +174,22 @@ impl<'source> Compiler<'source> {
             }
             ast::Stmt::EmitExpr(expr) => {
                 self.set_location_from_span(expr.span());
+
+                // detect {{ super() }} and {{ loop() }} as special instructions
+                if let ast::Expr::Call(call) = &expr.expr {
+                    if let ast::Expr::Var(var) = &call.expr {
+                        if var.id == "super" && call.args.is_empty() {
+                            self.add(Instruction::FastSuper);
+                            return Ok(());
+                        }
+                        if var.id == "loop" && call.args.len() == 1 {
+                            self.compile_expr(&call.args[0])?;
+                            self.add(Instruction::FastRecurse);
+                            return Ok(());
+                        }
+                    }
+                }
+
                 self.compile_expr(&expr.expr)?;
                 self.add(Instruction::Emit);
             }
@@ -181,7 +206,7 @@ impl<'source> Compiler<'source> {
                     // just outside of the loop.
                     self.add(Instruction::BuildList(0));
                     self.compile_expr(&for_loop.iter)?;
-                    self.start_for_loop(false);
+                    self.start_for_loop(false, false);
                     self.add(Instruction::DupTop);
                     self.compile_assignment(&for_loop.target)?;
                     self.compile_expr(filter_expr)?;
@@ -195,7 +220,7 @@ impl<'source> Compiler<'source> {
                     self.compile_expr(&for_loop.iter)?;
                 }
 
-                self.start_for_loop(true);
+                self.start_for_loop(true, for_loop.recursive);
                 self.compile_assignment(&for_loop.target)?;
                 for node in &for_loop.body {
                     self.compile_stmt(node)?;
@@ -323,6 +348,7 @@ impl<'source> Compiler<'source> {
                 });
             }
             ast::Expr::BinOp(c) => {
+                self.set_location_from_span(c.span());
                 let instr = match c.op {
                     ast::BinOpKind::Eq => Instruction::Eq,
                     ast::BinOpKind::Ne => Instruction::Ne,
@@ -342,14 +368,12 @@ impl<'source> Compiler<'source> {
                     ast::BinOpKind::Sub => Instruction::Sub,
                     ast::BinOpKind::Mul => Instruction::Mul,
                     ast::BinOpKind::Div => Instruction::Div,
-                    // TODO: floor div maps to normal div for now
-                    ast::BinOpKind::FloorDiv => Instruction::Div,
+                    ast::BinOpKind::FloorDiv => Instruction::IntDiv,
                     ast::BinOpKind::Rem => Instruction::Rem,
                     ast::BinOpKind::Pow => Instruction::Pow,
                     ast::BinOpKind::Concat => Instruction::StringConcat,
                     ast::BinOpKind::In => Instruction::In,
                 };
-                self.set_location_from_span(c.span());
                 self.compile_expr(&c.left)?;
                 self.compile_expr(&c.right)?;
                 self.add(instr);
